#include "rendering/pipelineText/data.h"
#include "rendering/pipelineText/pipeline.h"
#include "rendering/pipelineText/render.h"
#include <cglm/affine-pre.h>
#include <cglm/io.h>
#include <cglm/mat4.h>
#include <stdio.h>
#define STB_IMAGE_IMPLEMENTATION
#include <stb/stb_image.h>

#define STB_TRUETYPE_IMPLEMENTATION
#include <stb/stb_truetype.h>

#include "rendering/camera.h"
#include "rendering/init.h"
#include "rendering/vulkan/svulc.h"
#include "rendering/pipeline3d/setup.h"
#include "rendering/pipelineText/setup.h"
#include "rendering/rendering.h"

void fbResizeCb(GLFWwindow *window, int width, int height){
    struct rendering_windowUserPointer *wup = glfwGetWindowUserPointer(window);
    wup->s->shouldRecreateswapchainFlags = RNDR_RECREATE_SWAPCHAIN | RNDR_RECREATE_SWAPCHAIN_USE_TMP;
    wup->s->tmp_window_width = width;
    wup->s->tmp_window_height = height;
}

void keyCb(GLFWwindow *window, int key, int scancode, int action, int mods){
    struct rendering_windowUserPointer *wup = glfwGetWindowUserPointer(window); 
    struct camera *cam = wup->maincam;
    if(key == GLFW_KEY_ESCAPE && action == GLFW_PRESS){
        glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
    }
    if(key == GLFW_KEY_R && action == GLFW_PRESS){
        vec4 quat = {0.f,1.f,0.f,0.f};
        glm_quat_copy(quat, cam->rotation);
        glm_vec3_copy(GLM_VEC3_ZERO, cam->position);
    }
}

void keyboardUpdate(GLFWwindow *window, struct camera *cam){
    if(glfwGetKey(window, GLFW_KEY_W))
        glm_vec3_muladds(cam->forward, 0.1f, cam->position);
    if(glfwGetKey(window, GLFW_KEY_S))
        glm_vec3_muladds(cam->forward, -0.1f, cam->position);
    if(glfwGetKey(window, GLFW_KEY_A))
        glm_vec3_muladds(cam->right, -0.1f, cam->position);
    if(glfwGetKey(window, GLFW_KEY_D))
        glm_vec3_muladds(cam->right, 0.1f, cam->position);

    vec3 xaxis = {1.f,0.f,0.f};
    vec3 zaxis = {0.f,0.f,1.f};

    if(glfwGetKey(window, GLFW_KEY_UP))
        camera_angleAxis(cam, glm_rad(1.f), cam->right);
    if(glfwGetKey(window, GLFW_KEY_DOWN))
        camera_angleAxis(cam, glm_rad(-1.f), cam->right);
    if(glfwGetKey(window, GLFW_KEY_LEFT))
        camera_angleAxis(cam, glm_rad(1.f), cam->up);
    if(glfwGetKey(window, GLFW_KEY_RIGHT))
        camera_angleAxis(cam, glm_rad(-1.f), cam->up);

    camera_update(cam);
}

void cursorPosCb(GLFWwindow *window, double x, double y){
    struct rendering_windowUserPointer *wup = glfwGetWindowUserPointer(window); 
    struct camera *cam = wup->maincam;
    vec3 xaxis = {1.f,0.f,0.f};
    vec3 zaxis = {0.f,0.f,1.f};

    static double lastx;
    static double lasty;
    const double coef = 0.1f;
    glm_vec3_rotate(cam->rotation, glm_rad(x-lastx) * coef, zaxis);
    glm_vec3_rotate(cam->rotation, glm_rad(y-lasty) * coef, cam->right);
    lastx = x;
    lasty = y;
}

// callback that will be called when the programm encounters a fatal error so that it can be cleaned up
void die(){
    exit(EXIT_FAILURE);
}

int main(int argc, const char **argv){
    struct rendering_state s = {0};
    s.window_title = "Hi!";
    s.window_width = 600;
    s.window_height = 600;

    struct rendering_windowUserPointer wup = {
        NULL,
        &s
    };
    s.die = &die;

    glfwInitHint(GLFW_PLATFORM, GLFW_PLATFORM_X11);
    rendering_initGLFW(&s);

    glfwSetWindowUserPointer(s.window, &wup);
    glfwSetFramebufferSizeCallback(s.window, fbResizeCb);
    glfwSetKeyCallback(s.window, keyCb);
    glfwSetInputMode(s.window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);
    glfwSetCursorPosCallback(s.window, cursorPosCb);

    rendering_vkinstance(&s);
    printf("vkerror = %i\n", s.error);

    // TODO: make a function that selects the best, most suitable device
    unsigned int physDeviceNum;
    VkPhysicalDevice *physDevices;

    VkResult result = svlk_getPhysDeviceList(s.instance, &physDeviceNum, &physDevices);

    for (int i=0; i<physDeviceNum; i++) {
        if(svlk_deviceCompatable(physDevices[i], s.surface, VK_QUEUE_GRAPHICS_BIT, s.deviceextensions, s.numdeviceextensions)){
            s.physdevice = physDevices[i];
        }
    }
    // END TODO

    rendering_vkgetfamilies(&s);

    const char *deviceextensions[] = {VK_KHR_SWAPCHAIN_EXTENSION_NAME};
    s.deviceextensions = deviceextensions;
    s.numdeviceextensions = 1;
    rendering_vkdevice(&s);
    s.deviceextensions = 0;
    s.numdeviceextensions = 0;

    rendering_vkswapchain(&s);

    if(s.error != VK_SUCCESS) printf("fehler ): \n");

    // load a shader

    createPipeline3d(&s);
    createPipelineText(&s);

    rendering_createframebuffers(&s);

    rendering_createCommandPools(&s);
    rendering_createSyncObjects(&s);

    // create a triangle renderdata
    struct renderData rects[1] = {0};
    struct vertex rectverts[] = {
        {-0.5f, 0.f, 0.5f, 0.f, 0.f},
        {0.5f, 0.f, 0.5f, 1.f, 0.f},
        {-0.5f, 0.f, -0.5f, 0.f, 1.f},
        {0.5f, 0.f, -0.5f, 1.f, 1.f},
    };
    uint32_t triangleindices[] = {0,1,2, 1, 3, 2};

    renderData_create(&s, rectverts, 4, triangleindices, 6, &rects[0]);
    //triangles[1].position[2] = 2.f;
    // camera stuff
    uint32_t first = 1;

    struct camera cam;
    cam.far = 1000.f;
    cam.near = 0.1f;
    cam.fov = glm_rad(90.f);
    cam.position[1] = -2.f;
    cam.aspectRatio = 1.f;
    vec4 quat = {0.f,1.f,0.f,0.f};
    glm_quat_copy(quat, cam.rotation);
    camera_update(&cam);
    wup.maincam = &cam;

    // image loading
    int iw, ih, ic;
    stbi_uc *pixels = stbi_load("img.png", &iw, &ih, &ic, STBI_rgb_alpha);

    const unsigned char *fontdata = (const unsigned char*)readFileC("FuturaRenner-Regular.ttf");

    //stbtt_fontinfo fontinfo;
    //stbtt_InitFont(&fontinfo, fontdata, 0);
    
    void *fontpixels = malloc(512*512+1);
    void *chardata = malloc(256 * sizeof(stbtt_packedchar));

    stbtt_pack_context packingctx = {0};
    stbtt_PackBegin(&packingctx, fontpixels, 512, 512, 0, 1, NULL);

    stbtt_pack_range packingranges[] = {
        {
            64,
            0,
            NULL,
            256,
            chardata,
            0,
            0
        }
    };

    stbtt_PackFontRanges(&packingctx, fontdata, 0, packingranges, 1);

    stbtt_PackEnd(&packingctx);


    struct gpuimage img;
    if(fontpixels){
        rendering_createImageForRender(512, 512, fontpixels, 512 * 512 * 1, VK_FORMAT_R8_SRGB, &rects[0].albedo, &s);
        renderData_setupTextures(&s, &rects[0], NULL);
        img = rects[0].albedo;
        printf("IMG: avail!\n");
    } else {
        printf("IMG: not avail!\n");
        return 1;
    }

    struct Font ftrenner = {
        512,
        512,
        chardata,
        &img,
        NULL
    };
    setupFontTextureForRender(&s, &ftrenner, NULL);

    struct textElement textTest = {
        "Hallo es gibt noch kaum Satzzeichen! oder doch? häää",
        35,
        36,
        0,
        {{0}},
        &ftrenner,
        malloc(sizeof(struct PipelineTextUniformBufferObject) * 35)
    };

    glm_mat4_copy(GLM_MAT4_IDENTITY, textTest.matrix);

    vec3 mov = {-0.9,-0.9,0.0};
    glm_translate(textTest.matrix, mov);
    vec3 scale = {0.001,0.001,0.001};
    glm_scale(textTest.matrix, scale);

    pipelineText_submitTextElement(textTest, *textTest.font, textTest.compiled);

    while (!glfwWindowShouldClose(s.window)){
        glfwPollEvents();
        keyboardUpdate(s.window, &cam);

        struct ObjRenderArray rda = {
            rects,
            1,
            &textTest,
            1
        };
        
        render(&s, &s.renderTarget, &rda, 1, &cam);
        if(first){
            first = 0;
        }
        glm_vec3_print(cam.right, stdout);
        printf("\x1b[3A");
    }

}