#include "vulkan/gpumem.h"
#include "vulkan/svulc.h"
#include "vulkan/swapchain.h"
#include "vulkan/commandpools.h"
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <threads.h>
#include <unistd.h>
#include <vulkan/vulkan_core.h>
#include <GLFW/glfw3.h>
#include "../helpers.h"
#include "init.h"

VkShaderModule rendering_createShaderModule(sizedData shadercode, VkDevice device){
    VkShaderModuleCreateInfo createinfo = {
        VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
        NULL,
        0,
        shadercode.length,
        shadercode.data
    };
    VkShaderModule module;
    if(vkCreateShaderModule(device, &createinfo, NULL, &module)!= VK_SUCCESS){
        printf("Couldn't create shader module!");
        exit(EXIT_FAILURE);
    }
    return module;
}

void rendering_initGLFW(struct rendering_state *s){
    glfwInit();
    glfwWindowHint(GLFW_CLIENT_API, GLFW_NO_API);

    s->window = glfwCreateWindow(s->window_width, s->window_height, s->window_title, NULL, NULL);
    if(s->window == NULL){
        printf("oh noes, the window is null!\n");
    }

    s->extensions = glfwGetRequiredInstanceExtensions(&s->numextensions);
    if(s->numextensions == 0){
        printf("extensions is null!\n");
        const char *err;
        glfwGetError(&err);
        printf("error: %s\n", err);
    }
}

void rendering_vkinstance(struct rendering_state *s){
    const char *devExtensions[] = {VK_KHR_SWAPCHAIN_EXTENSION_NAME};
    const char *layers[] = {"VK_LAYER_KHRONOS_validation"};
    printf("exts (num: %i):\n", s->numextensions);
    for(int i = 0; i < s->numextensions; i++){
        printf("  ext:%s\n", s->extensions[i]);
    }
    checkErrorVk("vkCreateInstance",
        svlk_createIinstance(layers, 1, s->extensions, s->numextensions, &s->instance),
        s->die
    );

    checkErrorVk("glfwCreateWindowSurface",
        glfwCreateWindowSurface(s->instance, s->window, NULL, &s->surface),
        s->die
    );
}

void rendering_vkgetfamilies(struct rendering_state *s){
    int error = 0;
    s->queues.graphicsQFI = svlk_getQueueFamilyIndex(s->physdevice, VK_QUEUE_GRAPHICS_BIT, &error);
    if(error){return;}
    s->queues.presentationQFI = svlk_getPresentQueueFamilyIndex(s->physdevice, s->surface, &error);
    if(error){return;}
}

void rendering_vkdevice(struct rendering_state *s){
    float queuePriority = 1.f;

    VkDeviceQueueCreateInfo deviceQueues[] = {
        svlk_createQueueInfo(s->queues.graphicsQFI, 2, &queuePriority)
    };

    VkPhysicalDeviceFeatures devFeatures = {};

    VkPhysicalDeviceVulkan13Features dev13features = {
        VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES,
        NULL
    };
    dev13features.dynamicRendering = VK_TRUE;

    checkErrorVk("vkCreateDevice",
        svlk_createLogicalDevice(
            s->physdevice, s->surface, deviceQueues,1,
            devFeatures, s->deviceextensions,
            s->numdeviceextensions, &dev13features, &s->device
            ),
        s->die
    );

    vkGetDeviceQueue(s->device, s->queues.graphicsQFI, 0, &s->queues.graphicsQueue);
    vkGetDeviceQueue(s->device, s->queues.graphicsQFI, 0, &s->queues.transferQueue);
    vkGetDeviceQueue(s->device, s->queues.presentationQFI, 0, &s->queues.presentationQueue);
};

void rendering_vkswapchain(struct rendering_state *s){
    checkErrorVk("vkGetPhysicalDeviceSurfaceCapabilitiesKHR",
        vkGetPhysicalDeviceSurfaceCapabilitiesKHR(s->physdevice, s->surface, &s->surfaceCapabilities),
        s->die
    );
    
    s->surfaceformat = svlk_selectSwapchainFormat(s->physdevice, s->surface, VK_FORMAT_B8G8R8_SRGB, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR);

    s->swapchainExtent = svlk_checkSwapchainExtent(&s->surfaceCapabilities, s->window_width, s->window_height);
    //TODO
    s->swapchainNumimgs = s->surfaceCapabilities.minImageCount + 1;
    if(s->swapchainNumimgs > s->surfaceCapabilities.maxImageCount && s->surfaceCapabilities.maxImageCount != 0) s->swapchainNumimgs = s->surfaceCapabilities.maxImageCount;

    VkSwapchainCreateInfoKHR swcreateinfo = svlk_createSwapchainCreateinfo(
        s->swapchainNumimgs,
        s->surface,
        s->surfaceformat,
        VK_PRESENT_MODE_FIFO_KHR,
        s->swapchainExtent,
        1,
        VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT,
        s->surfaceCapabilities.currentTransform
    );

    unsigned int queueFamilies[] = {s->queues.graphicsQFI, s->queues.presentationQFI};
    svlk_selectSharingMode(&swcreateinfo, s->queues.graphicsQFI != s->queues.presentationQFI, queueFamilies, 2);

    checkErrorVk("vkCreateSwapchainKHR",
        vkCreateSwapchainKHR(s->device, &swcreateinfo, NULL, &s->swapchain),
        s->die
    );

    s->swapchainImages = svlk_getSwapchainImages(s->device, s->swapchain, &s->numswapchainImages);
    
    checkErrorVk("vkCreateImageView",
        svlk_createVkImageViews(s->device, s->swapchainImages, VK_IMAGE_VIEW_TYPE_2D, s->surfaceformat.format, s->numswapchainImages,&s->swapchainImageView),
        s->die
    );
    s->numswapchainImageView = s->numswapchainImages;
    
    // rendertarget
    rendering_createImage(s->window_width, s->window_height, 1, VK_IMAGE_TYPE_2D, s->surfaceformat.format, 1,
        VK_IMAGE_TILING_OPTIMAL, VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT,& s->renderTarget, s);
    svlk_createVkImageView(s->device, s->renderTarget.image, VK_IMAGE_VIEW_TYPE_2D, s->surfaceformat.format, &s->renderTarget.view);
}

void rendering_createframebuffers(struct rendering_state *s){
    s->framebuffers = calloc(s->numswapchainImageView, sizeof(VkFramebuffer));
    s->numframebuffers = s->numswapchainImageView;

    for(uint32_t i = 0; i < s->numswapchainImageView; i++){
        VkImageView attachments[] = {
            s->swapchainImageView[i]
        };

        VkFramebufferCreateInfo fbCI = {
            VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
            NULL,
            0,
            s->pipeline.renderpass,
            1,
            attachments,
            s->window_width,
            s->window_height,
            1
        };
        checkErrorVk("vkCreateFramebuffer",
            vkCreateFramebuffer(s->device, &fbCI, NULL, &s->framebuffers[i]),
            s->die
        );
    }
}
// IMPROTANT: also destroys rendertarget
void rendering_cleanupSwapchain(struct rendering_state *s){
    for(uint32_t i = 0; i < s->numframebuffers; i++){
        vkDestroyFramebuffer(s->device, s->framebuffers[i], NULL);
    }
    for(uint32_t i = 0; i < s->numswapchainImageView; i++){
        vkDestroyImageView(s->device, s->swapchainImageView[i], NULL);
    }
    rendering_destroyImage(&s->renderTarget, s);
    vkDestroySwapchainKHR(s->device, s->swapchain, NULL);
}

/* Here __was__ a relic of the old wor-- I mean relic of dumbness */

void rendering_createCommandPools(struct rendering_state *s){
    checkErrorVk("vkCreateCommandPool",
        svlk_createCommandPool(s->device, s->queues.graphicsQFI, VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT, &s->graphicsPool),
        s->die
    );
    checkErrorVk("vkAllocateCommandBuffer (graphicsBuffer)",
        svlk_allocateCommandBuffer(s->device, s->graphicsPool, VK_COMMAND_BUFFER_LEVEL_PRIMARY, 1, &s->graphicsBuffer),
        s->die
    );
    checkErrorVk("vkAllocateCommandBuffer (transferBuffer)",
        svlk_allocateCommandBuffer(s->device, s->graphicsPool, VK_COMMAND_BUFFER_LEVEL_PRIMARY, 1, &s->transferBuffer),
        s->die
    );
}

void rendering_createSyncObjects(struct rendering_state *s){
    checkErrorVk("vkCreateFence",
        svlk_createFence(s->device, &s->sync.frameFinishFence, VK_FENCE_CREATE_SIGNALED_BIT),
        s->die
    );
    checkErrorVk("vkCreateSemaphore",
        svlk_createSemaphore(s->device, &s->sync.presentReadySemaphore),
        s->die
    );
    checkErrorVk("vkCreateSemaphore",
        svlk_createSemaphore(s->device, &s->sync.renderFinishSemaphore),
        s->die
    );
}