#include "rendering.h"
#include "pipeline3d/render.h"
#include "pipeline3d/data.h"
#include <stdint.h>
#include <stdio.h>
#include <vulkan/vulkan_core.h>
#include "init.h"
#include "pipelineText/render.h"
#include "struct.h"
#include "vulkan/gpumem.h"
#include "vulkan/gpumem_types.h"
#include "renderobjs.h"

uint32_t render_getSwapchainPre(struct rendering_state *s, VkResult *res){
    // wait for next image
    uint32_t imageindex;
    VkResult nextImageResult = vkAcquireNextImageKHR(s->device, s->swapchain, UINT64_MAX, s->sync.presentReadySemaphore, NULL, &imageindex);

    if(res) *res = nextImageResult;
    
    // resize swapchain if surface size changed
    if(nextImageResult == VK_ERROR_OUT_OF_DATE_KHR){
        recreateSwapchain(s);
    }else if(nextImageResult != VK_SUCCESS && nextImageResult != VK_SUBOPTIMAL_KHR){
        printf("error getting next image!\n");
    }

    return imageindex;
}

void render_beginCmdBuf(struct rendering_state *s){ 
    // reset cmdbuf
    vkResetCommandBuffer(s->graphicsBuffer, 0);
   
    // begin new cmdbuf
    VkCommandBufferBeginInfo begininfo = {
        VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
        NULL,
        0,
        NULL
    };
    VkResult result = vkBeginCommandBuffer(s->graphicsBuffer, &begininfo);
}

void render_clearBg(struct rendering_state *s, struct gpuimage *target){
    // setup the rendertarget for the first clear
    rendering_transitionImageLayoutCB(s->graphicsBuffer, target->image, s->surfaceformat.format, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);

    // clear the rendertarget with clearcol
    VkImageSubresourceRange isubr = {
        VK_IMAGE_ASPECT_COLOR_BIT,
        0,
        1,
        0,
        1
    };

    VkClearColorValue clearcol = {0.1,0.5,0.2, 0.1f};

    vkCmdClearColorImage(s->graphicsBuffer, target->image, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, &clearcol, 1, &isubr);

    // transition the rendertarget to be a color attachment
    rendering_transitionImageLayoutCB(s->graphicsBuffer, target->image, s->surfaceformat.format, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL);
}

void render_copyToSwap(struct rendering_state *s, struct gpuimage *target, VkImage swimg){
    // setup the rendertarget and current swapchain image for the blit
    rendering_transitionImageLayoutCB(s->graphicsBuffer, target->image, s->surfaceformat.format, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL);
    rendering_transitionImageLayoutCB(s->graphicsBuffer, swimg, s->surfaceformat.format, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);

    // blit entire rendertarget to current swapchain image
    VkImageBlit blit = {
        {VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1},
        {{0,0,0}, {target->w, target->h, 1}},
        {VK_IMAGE_ASPECT_COLOR_BIT, 0, 0, 1},
        {{0,0,0}, {target->w, target->h, 1}},
    };

    vkCmdBlitImage(s->graphicsBuffer, target->image, VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, swimg, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &blit, VK_FILTER_NEAREST);

    // transition swapchain image back to be able to present
    rendering_transitionImageLayoutCB(s->graphicsBuffer, swimg, s->surfaceformat.format, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_PRESENT_SRC_KHR);

    vkEndCommandBuffer(s->graphicsBuffer);
}

void render_waitForDevice(struct rendering_state *s){
    // wait until device has finished rendering the previous frame
    vkWaitForFences(s->device, 1, &s->sync.frameFinishFence, 1, UINT32_MAX);
    // reset the fence to reuse it.
    vkResetFences(s->device, 1, &s->sync.frameFinishFence);
}

void render_submitToDevice(struct rendering_state *s){
    // submit the cmdbuf
    VkPipelineStageFlags waitstages[] = {VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT};
    VkSubmitInfo submitInfo = {
        VK_STRUCTURE_TYPE_SUBMIT_INFO,
        NULL,
        1,
        &s->sync.presentReadySemaphore,
        waitstages,
        1,
        &s->graphicsBuffer,
        1,
        &s->sync.renderFinishSemaphore
    };

    VkResult result = vkQueueSubmit(s->queues.graphicsQueue, 1, &submitInfo, s->sync.frameFinishFence);
}

void render_present(struct rendering_state *s, uint32_t imageindex, VkResult nextImageResult){
    // present the newly created image
    VkPresentInfoKHR presentI = {
        VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
        NULL,
        1,
        &s->sync.renderFinishSemaphore,
        1,
        &s->swapchain,
        &imageindex,
        NULL
    };

    vkQueuePresentKHR(s->queues.presentationQueue, &presentI);

    if(nextImageResult == VK_SUBOPTIMAL_KHR || s->shouldRecreateswapchainFlags){
        recreateSwapchain(s);
        s->shouldRecreateswapchainFlags = 0;
    }
}

void render(struct rendering_state *s, struct gpuimage *target, struct ObjRenderArray *rda, uint32_t objcount, struct camera *cam){
    render_waitForDevice(s);

    VkResult nextImageResult = 0;
    uint32_t imageindex = render_getSwapchainPre(s, &nextImageResult);

    render_beginCmdBuf(s);

    render_clearBg(s, target);

    VkExtent2D imageextent = {target->w, target->h};

    // do the rendering
    pipeline3d_recordCommandBuffer(
        s->graphicsBuffer,
        s->pipeline.renderpass,
        target->view,
        imageextent,
        VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
        s,
        &s->pipeline,
        rda->p3d_objs,
        rda->p3d_objs_count,
        cam);


    pipelineText_recordCommandBuffer(
        s->graphicsBuffer,
        s->pipelinetext.renderpass, 
        target->view, 
        imageextent, 
        VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL, 
        s, 
        &s->pipelinetext, 
        rda->pt_objs, 
        rda->pt_objs_count
    );

    render_copyToSwap(s, target, s->swapchainImages[imageindex]);

    render_submitToDevice(s);

    render_present(s, imageindex, nextImageResult);
}

void recreateSwapchain(struct rendering_state *s){
    if(s->shouldRecreateswapchainFlags & RNDR_RECREATE_SWAPCHAIN_USE_TMP){
        s->window_width = s->tmp_window_width;
        s->window_height = s->tmp_window_height;
    }

    vkDeviceWaitIdle(s->device);
    rendering_cleanupSwapchain(s);
    rendering_vkswapchain(s);
    rendering_createframebuffers(s);
}