#include "gpumem.h"
#include "gpumem_types.h"
#include "../../helpers.h"
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vulkan/vulkan_core.h>

void rendering_createBuffer(struct gpubuffer *buf, VkBufferUsageFlagBits usage, VkSharingMode shareMode, VkMemoryPropertyFlagBits properties, struct rendering_state *s){
    VkBufferCreateInfo bufferCI = {
        VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
        NULL,
        0,
        buf->size,
        usage,
        shareMode,
        0,
        NULL
    };

    vkCreateBuffer(s->device, &bufferCI, NULL, &buf->buffer);

    VkMemoryRequirements memreq;
    vkGetBufferMemoryRequirements(s->device, buf->buffer, &memreq);

    VkMemoryAllocateInfo allocinfo = {
        VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
        NULL,
        memreq.size,
        rendering_findMemoryType(memreq.memoryTypeBits, properties, s)
    };

    
    checkErrorVk("vkAllocateMemory", 
        vkAllocateMemory(s->device, &allocinfo, NULL, &buf->memory),
        s->die
    );
    checkErrorVk("vkBindBufferMemory", 
        vkBindBufferMemory(s->device, buf->buffer, buf->memory, 0),
        s->die
    );
}

void rendering_putBuffer(struct gpubuffer *buf, void *data, struct rendering_state *s){
    void *mapadr;
    vkMapMemory(s->device, buf->memory, 0, buf->size, 0, &mapadr);
    memcpy(mapadr, data, buf->size);
    vkUnmapMemory(s->device, buf->memory);
}

void rendering_makeBufferForRender(struct gpubuffer *buf, void *data, VkBufferUsageFlagBits usage, VkSharingMode shareMode, struct rendering_state *s){
    struct gpubuffer stagebuf = {};
    stagebuf.size = buf->size;
    rendering_createBuffer(&stagebuf, VK_BUFFER_USAGE_TRANSFER_SRC_BIT | usage, shareMode, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, s);
    rendering_putBuffer(&stagebuf, data, s);
    rendering_createBuffer(buf, VK_BUFFER_USAGE_TRANSFER_DST_BIT | usage, shareMode, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, s);
    rendering_copyBuffer(stagebuf.buffer, buf->buffer, buf->size, s);
    rendering_destroyBuffer(&stagebuf, s);
}

uint32_t rendering_findMemoryType(uint32_t type, VkMemoryPropertyFlags properties, struct rendering_state *s){
    VkPhysicalDeviceMemoryProperties memproperties;
    vkGetPhysicalDeviceMemoryProperties(s->physdevice, &memproperties);
    for(uint32_t i = 0; i < memproperties.memoryTypeCount; i++){
        if((type & (1 << i)) && (memproperties.memoryTypes[i].propertyFlags & properties) == properties){
            return i;
        }
    }
    printf("no suitable memory type!\n");
    exit(EXIT_FAILURE);
}

void rendering_endSubmitWait(VkCommandBuffer cmdbuf, VkQueue queue){
    vkEndCommandBuffer(cmdbuf);
    VkSubmitInfo submitinfo = {
        VK_STRUCTURE_TYPE_SUBMIT_INFO,
        NULL,
        0,
        NULL,
        0,
        1,
        &cmdbuf,
        0,
        NULL
    };
    vkQueueSubmit(queue, 1, &submitinfo, NULL);
    vkQueueWaitIdle(queue);
}

void rendering_copyBuffer(VkBuffer src, VkBuffer dst, VkDeviceSize size, struct rendering_state *s){
    VkCommandBufferBeginInfo begininfo = {
        VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
        NULL,
        0,
        NULL
    };
    vkBeginCommandBuffer(s->transferBuffer, &begininfo);
    VkBufferCopy copyregion = {
        0, 0,
        size
    };
    vkCmdCopyBuffer(s->transferBuffer, src, dst, 1, &copyregion);
    rendering_endSubmitWait(s->transferBuffer, s->queues.transferQueue);
}

void rendering_createImage(
    uint32_t w, uint32_t h, uint32_t d, VkImageType type, VkFormat fmt, uint32_t mipLevels, VkImageTiling tiling,
    VkImageUsageFlags usageflags, VkMemoryPropertyFlags properties, struct gpuimage *img, struct rendering_state *s){
    img->w = w;
    img->h = h;
    img->d = d;
    img->fmt = fmt;

    VkImageCreateInfo imgCI = {
        VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO,
        NULL,
        0,
        type,
        fmt,
        {w, h, d},
        mipLevels,
        1,
        VK_SAMPLE_COUNT_1_BIT,
        tiling,
        usageflags,
        VK_SHARING_MODE_EXCLUSIVE,
        0,
        NULL,
        VK_IMAGE_LAYOUT_UNDEFINED
    };
    checkErrorVk("vkCreateImage",
        vkCreateImage(s->device, &imgCI, NULL, &img->image),
        s->die
    );

    VkMemoryRequirements memreq;
    vkGetImageMemoryRequirements(s->device, img->image, &memreq);

    VkMemoryAllocateInfo allocinfo = {
        VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO,
        NULL,
        memreq.size,
        rendering_findMemoryType(memreq.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, s)
    };

    checkErrorVk("vkAllocateMemory", 
        vkAllocateMemory(s->device, &allocinfo, NULL, &img->memory),
        s->die
    );

    checkErrorVk("vkBindBufferMemory", 
        vkBindImageMemory(s->device, img->image, img->memory, 0),
        s->die
    );
}

void rendering_copyImageB(VkBuffer src, VkImage img, uint32_t w, uint32_t h, uint32_t d, VkImageAspectFlags aspect, struct rendering_state *s){
    VkCommandBufferBeginInfo begininfo = {
        VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
        NULL,
        0,
        NULL
    };
    vkBeginCommandBuffer(s->transferBuffer, &begininfo);

    VkBufferImageCopy region = {
        0,
        0,
        0,
        {aspect, 0, 0, 1},
        {0, 0, 0},
        {w, h, d}
    };
    vkCmdCopyBufferToImage(s->transferBuffer, src, img, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &region);

    rendering_endSubmitWait(s->transferBuffer, s->queues.transferQueue);
}

void rendering_transitionImageLayoutCB(VkCommandBuffer buf, VkImage img, VkFormat fmt, VkImageLayout old, VkImageLayout new){
    VkImageMemoryBarrier barrier = {
        VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,
        NULL,
        0,
        0,
        old,
        new,
        VK_QUEUE_FAMILY_IGNORED,
        VK_QUEUE_FAMILY_IGNORED,
        img,
        {
            VK_IMAGE_ASPECT_COLOR_BIT,
            0,
            1,
            0,
            1
        }
    };

    VkPipelineStageFlags srcStage;
    VkPipelineStageFlags dstStage;

    if(old == VK_IMAGE_LAYOUT_UNDEFINED && new == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL){
        barrier.srcAccessMask = 0;
        barrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;

        srcStage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
        dstStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
    } else if(old == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && new == VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL){
        barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;

        srcStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
        dstStage = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
    } else if(old == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && new == VK_IMAGE_LAYOUT_PRESENT_SRC_KHR){
        barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        barrier.dstAccessMask = 0;

        srcStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
        dstStage = VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
    } else if(old == VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL && new == VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL){
        barrier.srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
        barrier.dstAccessMask = VK_ACCESS_TRANSFER_READ_BIT;

        srcStage = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
        dstStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
    } else if(old == VK_IMAGE_LAYOUT_UNDEFINED && new == VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL){
        barrier.srcAccessMask = 0;
        barrier.dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
        
        srcStage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
        dstStage = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
    } else if(old == VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && new == VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL){
        barrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        barrier.dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
        
        srcStage = VK_PIPELINE_STAGE_TRANSFER_BIT;
        dstStage = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
    }

    vkCmdPipelineBarrier(buf, srcStage, dstStage, 0, 0, NULL, 0, NULL, 1, &barrier);
}
void rendering_transitionImageLayout(VkImage img, VkFormat fmt, VkImageLayout old, VkImageLayout new, struct rendering_state *s){
    VkCommandBufferBeginInfo begininfo = {
        VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
        NULL,
        0,
        NULL
    };
    vkBeginCommandBuffer(s->transferBuffer, &begininfo);

    rendering_transitionImageLayoutCB(s->transferBuffer, img, fmt, old, new);

    rendering_endSubmitWait(s->transferBuffer, s->queues.transferQueue);
}

void rendering_createImageForRender(uint32_t w, uint32_t h, void *data, uint64_t datalen, VkFormat fmt, struct gpuimage *img, struct rendering_state *s){
    struct gpubuffer stagebuf = {};
    stagebuf.size = datalen;
    img->size = datalen;
    // create and fill the staging buffer
    rendering_createBuffer(&stagebuf, VK_BUFFER_USAGE_TRANSFER_SRC_BIT, VK_SHARING_MODE_EXCLUSIVE, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT, s);
    rendering_putBuffer(&stagebuf, data, s);
    
    // create the image and transition the image layout
    rendering_createImage(w, h, 1, VK_IMAGE_TYPE_2D, fmt, 1, VK_IMAGE_TILING_OPTIMAL, VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT , VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, img, s);
    rendering_transitionImageLayout(img->image, fmt, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, s);

    // copy and transition to final stage
    rendering_copyImageB(stagebuf.buffer, img->image, w, h, 1, VK_IMAGE_ASPECT_COLOR_BIT, s);
    rendering_transitionImageLayout(img->image, fmt, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, s);

    rendering_destroyBuffer(&stagebuf, s);
};

void rendering_destroyImage(struct gpuimage *img, struct rendering_state *s){
    if(img->sampler) vkDestroySampler(s->device, img->sampler, NULL);
    if(img->view) vkDestroyImageView(s->device, img->view, NULL);
    vkDestroyImage(s->device, img->image, NULL);
    vkFreeMemory(s->device, img->memory, NULL);
    img->size = 0;
    img->w = 0;
    img->h = 0;
    img->d = 0;
    img->fmt = 0;
}