#ifndef __SWAPCHAIN_C__
#define __SWAPCHAIN_C__
#include "swapchain.h"
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <vulkan/vulkan_core.h>

VkSurfaceFormatKHR *svlk_getSurfaceFormats(VkPhysicalDevice physDevice, VkSurfaceKHR surface, unsigned int *len){
    vkGetPhysicalDeviceSurfaceFormatsKHR(physDevice, surface, len, NULL);

    if(len == 0) return NULL;

    VkSurfaceFormatKHR *ret = malloc(*len * sizeof(VkSurfaceFormatKHR));

    vkGetPhysicalDeviceSurfaceFormatsKHR(physDevice, surface, len, ret);
    return ret;
}

VkPresentModeKHR *svlk_getPresentModes(VkPhysicalDevice physDevice, VkSurfaceKHR surface, unsigned int *len){
    vkGetPhysicalDeviceSurfacePresentModesKHR(physDevice, surface, len, NULL);

    if(len == 0) return NULL;

    VkPresentModeKHR *ret = malloc(*len * sizeof(VkPresentModeKHR));

    vkGetPhysicalDeviceSurfacePresentModesKHR(physDevice, surface, len, ret);
    return ret;
}

VkSurfaceFormatKHR svlk_selectSwapchainFormatFromList(VkSurfaceFormatKHR *surfaceFormats, VkFormat format, VkColorSpaceKHR colorSpace, unsigned int *len){
    for (unsigned int i = 0; i < *len; i++) {
        if (surfaceFormats[i].format == format && surfaceFormats->colorSpace == colorSpace) {
            return surfaceFormats[i];
        }
    }
    return surfaceFormats[0];
}

VkSurfaceFormatKHR svlk_selectSwapchainFormat(VkPhysicalDevice physDevice, VkSurfaceKHR surface, VkFormat format, VkColorSpaceKHR colorSpace){
    unsigned int len = 0;
    VkSurfaceFormatKHR *surfaceFormats = svlk_getSurfaceFormats(physDevice, surface, &len);
    VkSurfaceFormatKHR surfaceFormat = svlk_selectSwapchainFormatFromList(surfaceFormats, format, colorSpace, &len);
    free(surfaceFormats);
    return surfaceFormat;
}

VkExtent2D svlk_checkSwapchainExtent(const VkSurfaceCapabilitiesKHR *capabilities, int w, int h){
    if (capabilities->currentExtent.width != UINT32_MAX) {
        return capabilities->currentExtent;
    } else {
        VkExtent2D extent = {
            w,
            h
        };
        if(w < capabilities->minImageExtent.width) extent.width = capabilities->minImageExtent.width;
        if(w > capabilities->maxImageExtent.width) extent.width = capabilities->maxImageExtent.width;

        if(h < capabilities->minImageExtent.height) extent.height = capabilities->minImageExtent.height;
        if(h > capabilities->maxImageExtent.height) extent.height = capabilities->maxImageExtent.height;

        return extent;
    }
}

VkSwapchainCreateInfoKHR svlk_createSwapchainCreateinfo(
    unsigned int imageCount,
    VkSurfaceKHR surface,
    VkSurfaceFormatKHR format,
    VkPresentModeKHR presentMode,
    VkExtent2D extent,
    int arrayLayers,
    unsigned int usage,
    VkSurfaceTransformFlagBitsKHR pretransform
    ){
    VkSwapchainCreateInfoKHR createinfo={0};
    createinfo.sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
    createinfo.surface = surface;
    createinfo.minImageCount = imageCount;
    createinfo.imageFormat = format.format;
    createinfo.imageColorSpace = format.colorSpace;
    createinfo.imageExtent = extent;
    createinfo.imageArrayLayers = arrayLayers;
    createinfo.imageUsage = usage;
    // createinfo.imageSharingMode  &co is handled in select image sharing mode
    createinfo.compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
    createinfo.presentMode = presentMode;
    createinfo.clipped = VK_TRUE;
    createinfo.oldSwapchain = NULL;
    createinfo.preTransform = pretransform;

    return createinfo;
}

void svlk_selectSharingMode(VkSwapchainCreateInfoKHR *createinfo, int concurrent, unsigned int *queueFamilies, int queueFamiliesNum){
    if(concurrent){
        createinfo->imageSharingMode = VK_SHARING_MODE_CONCURRENT;
        createinfo->queueFamilyIndexCount = queueFamiliesNum;
        createinfo->pQueueFamilyIndices = queueFamilies;
    } else {
        createinfo->imageSharingMode = VK_SHARING_MODE_EXCLUSIVE;
    }
}

VkImageViewCreateInfo svlk_createVkImageViewCreateInfo(VkImage image, VkImageViewType viewtype, VkFormat format){
    VkImageViewCreateInfo createInfo = {0};
    createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    createInfo.image = image;
    createInfo.viewType = viewtype;
    createInfo.format = format;
    createInfo.components.r = VK_COMPONENT_SWIZZLE_IDENTITY;
    createInfo.components.g = VK_COMPONENT_SWIZZLE_IDENTITY;
    createInfo.components.b = VK_COMPONENT_SWIZZLE_IDENTITY;
    createInfo.components.a = VK_COMPONENT_SWIZZLE_IDENTITY;
    createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
    createInfo.subresourceRange.baseMipLevel = 0;
    createInfo.subresourceRange.levelCount = 1;
    createInfo.subresourceRange.baseArrayLayer = 0;
    createInfo.subresourceRange.layerCount = 1;

    return createInfo;
}

VkImage *svlk_getSwapchainImages(VkDevice device, VkSwapchainKHR swapchain, unsigned int *imageCount){
    vkGetSwapchainImagesKHR(device, swapchain, imageCount, NULL);
    VkImage *images = malloc(*imageCount * sizeof(VkImage));
    vkGetSwapchainImagesKHR(device, swapchain, imageCount, images);
    return images;
}

VkResult svlk_createVkImageView(VkDevice device, VkImage image, VkImageViewType viewtype, VkFormat format, VkImageView *imageView){
    VkImageViewCreateInfo createInfo = svlk_createVkImageViewCreateInfo(image, viewtype, format);

    return vkCreateImageView(device, &createInfo, NULL, imageView);
}

VkResult svlk_createVkImageViews(VkDevice device, VkImage *images, VkImageViewType viewtype, VkFormat format, unsigned int numimgs, VkImageView **imageView){
    *imageView = malloc(numimgs * sizeof(VkImageView));
    VkResult result;
    for(int i = 0; i < numimgs; i++){
        result = svlk_createVkImageView(device, images[i], viewtype, format, &(*imageView)[i]);
        if(result != VK_SUCCESS) return result;
    }
    return VK_SUCCESS;
}

#endif